/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include "CM_util.h"
#include "CM_lse.h"


const string trainingFileName = "train.data";
const string validationFileName = "valid.data";
const string std_correctionFileName = "stdev.data";
const string eofVectorsFileName = "eofVectors.data";
const string eofSingularsFileName = "eofSingulars.data";
const int nvar = 1;
const int BUF_SIZE = 1024*1024*4;
const long seed = 42;
const string arFileName = "model.values";

using namespace CheapMatrix;

class ARModeler {
protected:
    Matrix coefs;
    int order;
public:
    /*
    Fitting is done like this:

    A =
        y(e,arorder)    y(e,arorder-1)  ...   y(e,1)
        y(e,arorder+1)  y(e,arorder)    ...   y(e,2)
        ...             ...             ...   ...
        y(e,max_time-1) y(e,max_time-2) ...   y(e,max_time-arorder)

    b =
        y(e,arorder+1)
        y(e,arorder+2)
        ...
        y(e,max_time)

    For each eof component e.

    With y(i,t) = x(i,t) - mean, x(i,t) beeing the ith component of the
    data in the eof basis at time t. mean is the average in this basis
    of the component vectors.
    A least square is then done, A * alpha = b, to find the coefficients
    of the AR model Y(t) = a1.Y(t-1) + a2.Y(t-2) + ... + aOrder.Y(t-Order)
    each aj is the subvector of alpha with the corresponding components
    for the data in the EOF basis.

    Fitting is done like this in vector mode, that is:
    Y(t) = a1.Y(t-1) + a2.Y(t-2) + ... + aOrder.Y(t-Order)
    with ai for the whole vector

    A =
        y(1,arorder)    y(1,arorder-1)        ... y(1,1)
        y(2,arorder)    y(2,arorder-1)        ... y(2,1)
        ...             ...                   ... ...
        y(neof,arorder) y(neof,arorder-1)     ... y(neof,1)
        y(1,arorder+1)  y(1,arorder)          ... y(1,2)
        y(2,arorder+1)  y(2,arorder)          ... y(2,2)
        ...             ...                   ... ...
        y(neof,arorder+1) y(neof,arorder)     ... y(neof,2)
        ...
        y(neof,max_time-1) y(neof,max_time-2) ... y(neof,max_time-arorder)

    b =
        y(1,arorder+1)
        y(2,arorder+1)
        ...
        y(neof,arorder+1)
        y(1,arorder+2)
        y(2,arorder+2)
        ...
        y(neof,arorder+2)
        ...
        y(neof,max_time)
    */
    ARModeler(int order);
    bool fitcoef(const Matrix &series);
    bool fitvector(const Matrix &series);
    Vector predict(const Matrix &series);
    // Can save the model in a file
    bool save(const string& filename);
    // Can load the model from a file
    bool load(const string& filename);
};

ARModeler::ARModeler(int _order) : order(_order)
{
}

bool ARModeler::fitcoef(const Matrix &series)
{
    int n = series.cols();
    if (order>=n) {
        cerr << "AR order (" << order << ") must be less than data number (" << n << ")!" << endl;
        return false;
    }
    int m = series.rows();
    coefs = zeros(m,order);
    // Use all-in-one-memory-block temporary matrices for LAPACK
    Matrix A = Matrix::temporary(n-order, order);
    Vector B = Vector::temporary(n-order);
    Vector ealpha(order);
    for (int e=1; e<=m; e++) {
        for (int j=1; j<=order; j++) for (int i=1; i<=(n-order); i++)
            A(i,j) = series(e,order+i-j);
        for (int i=1; i<=(n-order); i++) B(i) = series(e,order+i);
        if (!LSE(A,B,ealpha)) {
            cerr << "ARCoefFit: Least square computation failed." << endl;
            return false;
        }
        for (int i=1; i<=order; i++) coefs(e,i) = ealpha(i);
    }
    return true;
}

bool ARModeler::fitvector(const Matrix &series)
{
    int n = series.cols();
    if (order>=n) {
        cerr << "AR order (" << order << ") must be less than data number (" << n << ")!" << endl;
        return false;
    }
    int m = series.rows();
    coefs = zeros(1,order);
    // First generate a big matrix with the values duplicated for LAPACK
    Matrix A = Matrix::temporary(m*(n-order), order);
    for (int j=0; j<order; j++) for (int i=0; i<(n-order); i++)
        A(i*m+1, (i+1)*m, j+1, j+1).copy(series(1,m,order-j+i,order-j+i));
    // use existing Matrix, with reshaping, for target values
    if (!LSE(A, series(1,m,order+1,n).reshape(m*(n-order),1), coefs.reshape(order,1))) {
        cerr << "ARVectorFit: Least square computation failed." << endl;
        return false;
    }
    return true;
}

Vector ARModeler::predict(const Matrix &series)
{
    if (order!=series.cols()) {
        cerr << "ARPredict: The series must contain the AR order (" << order << ") column vectors." << endl;
        return Vector();
    }
    int m = series.rows();
    int model_dim = coefs.rows();
    if ((model_dim>1) && (model_dim!=m)) {
        cerr << "ARPredict: AR model dimension (" << model_dim << ") and data dimension (" << m << ") do not match!" << endl;
        return Vector();
    }
    Vector predicted = zeros(m);
    if (model_dim>1) {
        for (int k=1; k<=order; k++)
            predicted += multiply(coefs(1,m,k,k), series(1,m,order-k+1,order-k+1));
    } else {
        for (int k=1; k<=order; k++)
            predicted += coefs(1,k) * series(1,m,order-k+1,order-k+1);
    }
    return predicted;
}

// Can save the model in a file
bool ARModeler::save(const string& filename)
{
    if (filename.empty()) return false;
    return ::CheapMatrix::save(coefs, filename);
}

// Can load the model from a file
bool ARModeler::load(const string& filename)
{
    if (filename.empty()) return false;
    Matrix loadedValues;
    if (!::CheapMatrix::load(loadedValues, filename)) return false;
    if (loadedValues.cols()!=order) return false;
    coefs = loadedValues;
    return true;
}

int help(int argc, char** argv)
{
    cout <<
argv[0] << ": Fits an autoregressive model.\n"
"  Usage: " << argv[0] << " lag=order max_step_ahead v|c [log_eof_num...]\n"
"The AR model is loaded before or saved after the fitting "
"in the file '" << arFileName << "', if possible.\n"
"The lag argument is mandatory.\n"
"max_step_ahead is the maxium number of step ahead for which the prediction "
"RMSE will be computed.\n"
"The next argument determines if the model should have one coefficient per "
"vector (v) or one per coefficient (c). The last is equivalent to running "
"as many 1D AR-models as coefficients in parallel.\n"
"This argument is not significant when loading an existing model.\n"
"It is also possible to log the predicted series associated with the EOF "
"numbers given in argument. On each line of the log file will be the "
"corresponding step-ahead values, the series being column-wise.\n"
    ;
    return 1;
}
#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
    cerr << msg << " " << arg << endl; \
    return 2; \
}

int main(int argc, char** argv)
{
    time_t beginTime = time(0);
    // we stay in the background when losing connection to the terminal
    signal(SIGHUP, SIG_IGN); // not necessary with zsh, but for the others...

    const int mandatory_args = 3;
    if (argc<=mandatory_args) return help(argc, argv);
    process_arg(int, lag, atoi, argv[1], lag<1, "Invalid lag")
    process_arg(int, max_step_ahead, atoi, argv[2], max_step_ahead<1, "Invalid max step ahead prediction limit")
    const bool vectormode = !strcmp(argv[3],"v");
    if (!vectormode && strcmp(argv[3],"c")) {
        cerr << "Invalid mode " << argv[3] << ". This must be 'c', or 'v'." << endl;
        return 2;
    }

    randinit(seed); // have reproducible results

    ofstream output;
    output.precision(10);
    output.setf(ios::scientific);

    cout << "Loading data." << endl;
    Matrix training, validation, eofVectors;
    Vector std_correction, eofSingulars;
    if (!load(eofVectors,eofVectorsFileName)) {
        cerr << "Could not load eofVectors from '" << eofVectorsFileName << "'." << endl;
        return 2;
    } else cout << "eofVectors loaded from '" << eofVectorsFileName << "'." << endl;
    if (!load(eofSingulars,eofSingularsFileName)) {
        cerr << "Could not load eofSingulars from '" << eofSingularsFileName << "'." << endl;
        return 2;
    } else cout << "eofSingulars loaded from '" << eofSingularsFileName << "'." << endl;
    const int neof = eofVectors.cols();
    if (!neof || neof != eofSingulars.rows()) {
        cerr << "Invalid number of EOFs in eofVectors or eofSingulars." << endl;
        return 2;
    }
    const int m = eofVectors.rows();
    cout << "Data dimension is " << m << ", number of variables=" << nvar << ", number of EOFs=" << neof << endl;
    if (!load(std_correction, std_correctionFileName)) {
        cerr << "Could not std_correction from '" << std_correctionFileName << "'." << endl;
        return 2;
    } else cout << "std_correction loaded from '" << std_correctionFileName << "'." << endl;
    if (m != std_correction.rows()) {
        cerr << "std_correction dimension does not match EOFs dimension." << endl;
        return 2;
    }
    if (!load(training, trainingFileName)) {
        cerr << "Could not training from '" << trainingFileName << "'." << endl;
        return 2;
    } else cout << "training loaded from '" << trainingFileName << "'." << endl;
    if (neof != training.rows()) {
        cerr << "Training data dimension does not match number of EOFs." << endl;
        return 2;
    }
    const int t = training.cols();
    cout << "Training data size is " << t << " elements."<< endl;
    if (!load(validation, validationFileName)) {
        cerr << "Could not validation from '" << validationFileName << "'." << endl;
        return 2;
    } else cout << "validation loaded from '" << validationFileName << "'." << endl;
    if (neof != validation.rows()) {
        cerr << "Validation data dimension does not match number of EOFs." << endl;
        return 2;
    }
    const int v = validation.cols();
    cout << "Validation data size is " << v << " elements."<< endl;
    const int var_size = m / nvar;

    int nlogeof = 0;
    int *loggedeofnum = 0;
    if (argc>mandatory_args+1 && strcmp(argv[mandatory_args+1],"sanity")) {
        loggedeofnum = new int[argc-mandatory_args-1];
        for (int i=mandatory_args+1; i<argc; i++) {
            int num = atoi(argv[i]);
            if (num <= 0 || num > neof) {
                cerr << "Skipping invalid EOF number " << argv[i] << endl;
                continue;
            } else {
                cout << "EOF number " << num << " will be logged." << endl;
            }
            loggedeofnum[nlogeof++] = num;
        }
    }

    cout << "Creating AR-" << lag << " model." <<endl;
    // Create the autoencoder
    ARModeler armodel(lag);

    // Try to load the model first
    if (armodel.load(arFileName))
        cout << "Model loaded from '" << arFileName << "'." << endl;
    else {
        cout << "Could not load model from '" << arFileName << "'. Fitting a new one." << endl;
        // Fit it
        if (vectormode) armodel.fitvector(training);
        else armodel.fitcoef(training);
        // Save it
        if (armodel.save(arFileName))
            cout << "Model saved in '" << arFileName << "'." << endl;
        else cout << "Could not save model in '" << arFileName << "'." << endl;
    }

    // Test on the validation set
    cout << "Testing Model" <<endl;
    ofstream logfile[nlogeof];
    char varfile[100];
    for (int l = 0; l<nlogeof; l++) {
        sprintf(varfile,"ar%d_forecast_eof%d_predicted_series.txt",lag,loggedeofnum[l]);
        logfile[l].open(varfile);
        logfile[l].precision(10);
        logfile[l].setf(ios::scientific);
    }
    Matrix error = zeros(m,max_step_ahead);
    Matrix valid_block(neof,lag);
    for (int i=1; i<=v-lag-max_step_ahead+1; i++) {
        valid_block.copy(validation(1,neof,i,i+lag-1));
        for (int s=1; s<=max_step_ahead; s++) {
            Vector prediction = armodel.predict(valid_block);
            for (int l = 0; l<nlogeof; l++)
                logfile[l] << prediction(loggedeofnum[l]) << " ";
            // Reconstruct the data from the EOFs
            Vector eof_error = prediction - validation(1,neof,i+lag+s-1,i+lag+s-1);
            Vector rec_error = multiply(
                eofVectors * multiply(eofSingulars, eof_error),
                std_correction);
            // Add this to the total error
            error(1,m,s,s) += multiply(rec_error,rec_error);
            // update valid_block with the new value (if necessary)
            if (s<max_step_ahead) {
                valid_block(1,neof,1,lag-1).copy(valid_block(1,neof,2,lag));
                valid_block(1,neof,lag,lag).copy(prediction);
            }
        }
        for (int l = 0; l<nlogeof; l++) logfile[l] << endl;
        cout << "." << flush;
        if (!(i%10)) cout << " " << flush;
        if (!(i%50)) cout << i << "/" << v-lag-max_step_ahead+1 << endl;
    }
    cout << " " << v-lag-max_step_ahead+1 << "/" << v-lag-max_step_ahead+1 << endl;
    // Close log files
    for (int l = 0; l<nlogeof; l++) logfile[l].close();
    // Error is MSE for now. Have to average on variables _before_ taking
    // the square root.
    error /= v-lag-max_step_ahead+1;

    // RMSE calculation
    cout << "Calculating RMSE" <<endl;
    for (int var = 0; var < nvar; var++) {
        double rmse = 0;
        cout << "Variable " << var+1 << ": " << endl;
        char varfile[100];
        sprintf(varfile,"ar%d_forecast_RMSE_var%d.txt",lag,var+1);
        output.open(varfile);
        for (int s=1; s<=max_step_ahead; s++) {
            rmse = sqrt( mean(error(var*var_size+1, (var+1)*var_size, s, s)) );
            cout << s << " step ahead: " << rmse << endl;
            output << rmse << " " << endl;
        }
        output << endl; output.close();
    }

    cout << "Total time " << time(0) - beginTime << " seconds." << endl;
    return 0;
}

