/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include "CM_util.h"
#include "CM_optimize.h"

const char* validationFileName = "valid.data";
const char* std_correctionFileName = "stdev.data";
const char* eofVectorsFileName = "eofVectors.data";
const char* eofSingularsFileName = "eofSingulars.data";
const int nvar = 1;
const int BUF_SIZE = 1024*1024*4;
const long seed = 42;

using namespace CheapMatrix;
int help(int argc, char** argv)
{
    cout << "Climatology prediction." << endl;
    return 1;
}

#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
    cerr << msg << " " << arg << endl; \
    return 2; \
}

int main(int argc, char** argv)
{
    time_t beginTime = time(0);
    // we stay in the background when losing connection to the terminal
    signal(SIGHUP, SIG_IGN); // not necessary with zsh, but for the others...

    ofstream output;
    output.precision(10);
    output.setf(ios::scientific);

    cout << "Loading data." << endl;
    Matrix training, validation, eofVectors;
    Vector std_correction, eofSingulars;
    if (!load(eofVectors,eofVectorsFileName)) {
        cerr << "Could not load eofVectors from '" << eofVectorsFileName << "'." << endl;
        return 2;
    } else cout << "eofVectors loaded from '" << eofVectorsFileName << "'." << endl;
    if (!load(eofSingulars,eofSingularsFileName)) {
        cerr << "Could not load eofSingulars from '" << eofSingularsFileName << "'." << endl;
        return 2;
    } else cout << "eofSingulars loaded from '" << eofSingularsFileName << "'." << endl;
    const int neof = eofVectors.cols();
    if (!neof || neof != eofSingulars.rows()) {
        cerr << "Invalid number of EOFs in eofVectors or eofSingulars." << endl;
        return 2;
    }
    const int m = eofVectors.rows();
    cout << "Data dimension is " << m << ", number of variables=" << nvar << ", number of EOFs=" << neof << endl;
    if (!load(std_correction, std_correctionFileName)) {
        cerr << "Could not std_correction from '" << std_correctionFileName << "'." << endl;
        return 2;
    } else cout << "std_correction loaded from '" << std_correctionFileName << "'." << endl;
    if (m != std_correction.rows()) {
        cerr << "std_correction dimension does not match EOFs dimension." << endl;
        return 2;
    }
    if (!load(validation, validationFileName)) {
        cerr << "Could not validation from '" << validationFileName << "'." << endl;
        return 2;
    } else cout << "validation loaded from '" << validationFileName << "'." << endl;
    if (neof != validation.rows()) {
        cerr << "Validation data dimension does not match number of EOFs." << endl;
        return 2;
    }
    const int v = validation.cols();
    cout << "Validation data size is " << v << " elements."<< endl;
    const int var_size = m / nvar;


    // Test on the validation set
    cout << "Computing climatology predictions." <<endl;
    Vector error = zeros(m);
    Vector prediction = zeros(neof,1);
    for (int i=1; i<=v; i++) {
        // Reconstruct the data from the EOFs
        Vector eof_error = prediction - validation(1,neof,i,i);
        Vector rec_error = multiply(
                eofVectors * multiply(eofSingulars, eof_error),
                std_correction);
        // Add this to the total error
        error(1,m) += multiply(rec_error,rec_error);
    }
    // Error is MSE for now. Have to average on variables _before_ taking
    // the square root.
    error /= v;

    // RMSE calculation
    cout << "Calculating RMSE" <<endl;
    for (int var = 0; var < nvar; var++) {
        double rmse = 0;
        cout << "Variable " << var+1 << ": " << flush;
        char varfile[100];
        sprintf(varfile,"climatology_RMSE_var%d.txt",var+1);
        output.open(varfile);
        rmse = sqrt( mean(error(var*var_size+1, (var+1)*var_size)) );
        cout << rmse << endl;
        output << rmse << endl;
        output.close();
    }

    cout << "Total time " << time(0) - beginTime << " seconds." << endl;
    return 0;
}

