/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "CM_util.h"
#include "CM_svd.h"
#include "CM_optimize.h"

const string dataFileName = "Z500_85_94_5degrees.data";
const string meanFileName = "mean.data";
const string sineFileName = "seasonal_sines.data";
const string stdevFileName = "stdev.data";
const string eofVectorsFileName = "all_eofVectors.data";
const string eofSingularsFileName = "all_eofSingulars.data";
const int m = 320, n = 3346, t = 1826, v = 1520, nvar = 1;
const int var_size = m / nvar;
const int BUF_SIZE = 1024*1024;
const int seed = 42;

using namespace CheapMatrix;

// Model to fit a sine with the given period to a data vector
// The model is of the form S(t) = alpha * sin (omega * t + phi);
class SineModel : public OptimizeTarget
{
protected:
	double omega;
	// Evaluate error and gradients on this series
	MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH);
public:
	SineModel(double _omega);
};

SineModel::SineModel(double _omega) : omega(_omega)
{
	// Set initial values to something random
	values = randn(2);
	gradients = zeros(2);
}

// Evaluate error and gradients on this series
MatrixType SineModel::evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what)
{
	Vector common = omega * data + values(2);
	Vector common2 = sin(common);
	Vector common3 = values(1) * common2 - target;
	if (what & OptimizeTarget::EVALUATE_GRADIENT) {
		gradients(1) = dot(common3, common2);   // alpha derivative
		gradients(2) = dot(common3, values(1) * cos(common)); // phi derivative
	}
	if (what & OptimizeTarget::EVALUATE_ERROR) {
		return 0.5 * dot(common3, common3);  // error
	} else return 0;
}

int help(int argc, char** argv)
{
	cout <<
"This generates the series from the data.\n"
"  Usage: " << argv[0] << " day_freq\n"
"\n"
"The data is first mean corrected, and a seasonal sine (fixed period of "
"2 * pi / 365.25) is fitted to each variable.\n"
"This component is then also removed, and the data normalized by "
"its standard deviation.\n"
"EOFs are computed using only one every day_freq days.\n"
	;
	return 1;
}
#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
	cerr << msg << " " << arg << endl; \
	return 2; \
}

int main(int argc, char** argv)
{
	time_t beginTime = time(0);
	// Stay alive if parent is killed (as for connection lost)
	signal(SIGHUP, SIG_IGN);

	
/*	// Subsampling
{	cout << "Loading data." << endl;
	Matrix data(1280, 3346);
	if (!load(data,dataFileName) || data.cols()!=n || data.rows()!=m) {
		cerr << "Could not load data from '" << dataFileName << "'." << endl;
		return 2;
	} else cout << "Data loaded from '" << dataFileName << "'." << endl;
	cout << "Subsampling" << endl;
	Matrix subs(320, 3346);
	for (int i=1; i<=3346; i++) subs(all,i).copy(data(all,i).reshape(16,80)(range(1,2,16),all)(all,range(1,2,80)).reshape(320,1));
	if (save(subs, "subsampled.data")) cout << "OK" << endl;
	else cout << "error" << endl;
	return 0;
}*/	
	
	
	
	
	const int mandatory_args = 1;
	if (argc<=mandatory_args) return help(argc, argv);
	process_arg(int, day_freq, atoi, argv[1], day_freq<1, "Invalid day frequency")

	cout << "Loading data." << endl;
	Matrix data(m,n);//, BUF_SIZE, "data.fmat",true);
	if (!load(data,dataFileName) || data.cols()!=n || data.rows()!=m) {
		cerr << "Could not load data from '" << dataFileName << "'." << endl;
		return 2;
	} else cout << "Data loaded from '" << dataFileName << "'." << endl;
	assert(t<n);
	int mindim = (int)min(m,(t-1)/day_freq+1); // max number of EOF vectors
	assert(t+v==n);
	Matrix training = data(all,1,t);

	cout << "Computing mean correction." << endl;
	Vector mean_correction = mean(training,2); // mean along dim 2
	if (save(mean_correction, meanFileName))
		cout << "mean_correction saved in '" << meanFileName << "'." << endl;
	else {cerr << "Could not save mean_correction!" << endl; return 2;}

	cout << "Removing seasonality." << endl;
	randinit(seed);
	// PI isn't present on all systems
	// I should really declare some useful constants (e, pi, Inf, NaN, eps...)
	// in the namespace
	#ifndef M_PI
	#define M_PI 3.14159265358979323846
	#endif
	const double omega = 2.0*M_PI/365.25;
	SineModel sine(omega);
	//cout << sine.gradcheck(range(1,t), training(1,all)) << endl; return 0;
	SCG scg;
	scg.option(1) = 100;
	// scg.log(&cerr);    // display this lot on another output
	Matrix params(2,m);
	for (int i=1; i<=m; i++) {
		scg.optimize(sine, range(1,t), training(i,all).transpose()-mean_correction(i));
		Vector alphi = sine.copy_values();
		// Correct all the data (extand to the validation set as well)
		data(i,all) -= alphi(1)*sin(omega*range(1,n).transpose()+alphi(2)) + mean_correction(i);
		params(all,i).copy(alphi);
		cout << "." << flush;
		if (!(i%10)) cout << " " << flush;
		if (!(i%50)) cout << i << endl;
	}
	cout << " " << m << endl;
	if (save(params, sineFileName))
		cout << "Seasonal sine parameters (amplitude, phase) saved in '" << sineFileName << "'." << endl;
	else {cerr << "Could not save sine parameters!" << endl; return 2;}

	cout << "Standard deviation correction." << endl;
	Vector stdev_correction = stdev(training,2); // std along dim 2
	for (int j=1; j<=n; j++) data(all,j).divide(stdev_correction);
	if (save(stdev_correction, stdevFileName))
		cout << "stdev_correction saved in '" << stdevFileName << "'." << endl;
	else {cerr << "Could not save stdev_correction!" << endl; return 2;}

	cout << "Computing EOFs." << endl;
	// EOFs on the training set: nowhere the validation set is touched
	Matrix eofVectors(m, mindim);//,BUF_SIZE, "eofVectors.fmat",true);
	Vector eofSingulars(mindim);// in memory
	if (!SVD(training(all, range(1,day_freq,t)), eofVectors, eofSingulars)) {
		cerr << "Error computing EOFs." << endl;
		return 2;
	}
	if (save(eofVectors, eofVectorsFileName))
		cout << "EOF Vectors saved in '" << eofVectorsFileName << "'." << endl;
	else {cerr << "Could not save eofVectors!" << endl; return 2;}
	if (save(eofSingulars, eofSingularsFileName))
		cout << "EOF Singulars saved in '" << eofSingularsFileName << "'." << endl;
	else {cerr << "Could not save eofSingulars!" << endl; return 2;}

	cout << "Total time " << time(0) - beginTime << " seconds." << endl;
	return 0;
}
