/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include "CM_util.h"
#include "CM_optimize.h"

const string trainingFileName = "train.data";
const string validationFileName = "valid.data";
const string std_correctionFileName = "stdev.data";
const string eofVectorsFileName = "eofVectors.data";
const string eofSingularsFileName = "eofSingulars.data";
const int nvar = 1;
const int BUF_SIZE = 1024*1024*4;
const long seed = 42;
const string netFileName = "network.values";

using namespace CheapMatrix;

// Multi-layer perceptron
// TODO: Add output activation functions other than linear
class MLP : public OptimizeTarget {
protected:
	Matrix *W; // weight array (will be references to values)
	Vector *B; // bias array (will be references to values)
	Matrix *gW; // weight array (will be references to gradients)
	Vector *gB; // bias array (will be references to gradients)
	Vector *H; // neuron values of the last forward
	int nlayers; // number layers
	int input_dim, output_dim, weight_dim, total_dim;
	double prior;
	// Evaluate error and gradients on the data
	// uses batch backpropagation
	MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH);
public:
	// MLP stuff
	//  ............ Input
	//   \||||||||/  W[1]
	//    ........   + B[1] = layer 1
	//    \|||||/    W[2]
	//     .....     + B[2] = layer 2
	//    /|||||\    W[3]
	//    ........   + B[3] = layer 3
	//   /||||||||\  W[4]
	//  ............ + B[4] = Data data t
	//
	// Arguments :
	//  1. number of layers (including input/output layers)
	//  2. prior on the weights for the weight decay. Equivalent to the
	//     inverse of the variance presumed on the weights. Can be 0 to
	//     ignore weight decay.
	//  3+. input dim, hidden1 dim, hidden2 dim... ouput dim
	MLP(int _nlayers, double _prior, ...);
	~MLP();
	// Forward the input data, which must be of the input dimension
	// Returns a vector of the output dimension
	virtual Vector forward(const Vector& input);
};

MLP::MLP(int _nlayers, double _prior, ...) : nlayers(_nlayers), prior(_prior)
{
	assert(_nlayers > 2 ); // At least one hidden layer
	int *dim = new int[_nlayers];
	va_list ap;
	va_start(ap,_prior);
	for (int i=0; i<_nlayers; i++) dim[i] = va_arg(ap, int);
	va_end(ap);
	int *lim = new int[(_nlayers-1)*2+1];
	lim[0] = 0;
	for (int i=1; i<_nlayers; i++) {
		lim[i] = lim[i-1] + dim[i] * dim[i-1];
	}
	for (int i=0; i<_nlayers-1; i++) {
		lim[i+_nlayers] = lim[i+_nlayers-1] + dim[i+1];
	}
	weight_dim = lim[_nlayers-1];
	total_dim = lim[(_nlayers-1)*2];
	input_dim = dim[0];
	output_dim = dim[_nlayers-1];
	values = randn(total_dim);    // Set initial values to something random
	gradients = zeros(total_dim); // And gradients to 0
	// Set the references right
	W = new Matrix[_nlayers-1];
	B = new Vector[_nlayers-1];
	gW = new Matrix[_nlayers-1];
	gB = new Vector[_nlayers-1];
	H = new Vector[_nlayers-2];
	for (int i=0; i<_nlayers-1; i++) {
		W[i] = values(lim[i]+1,lim[i+1]).reshape(dim[i+1],dim[i]);
		gW[i] = gradients(lim[i]+1,lim[i+1]).reshape(dim[i+1],dim[i]);
		B[i] = values(lim[i+_nlayers-1]+1,lim[i+_nlayers]);
		gB[i] = gradients(lim[i+_nlayers-1]+1,lim[i+_nlayers]);
		W[i] /= sqrt(dim[i] + 1);
		B[i] /= sqrt(dim[i + 1]);
	}
	delete lim; delete dim;
}

MLP::~MLP()
{
	delete W; delete gW;
	delete B; delete gB;
	delete H;
}

Vector MLP::forward(const Vector& input)
{
	assert(input.rows() == input_dim);
	H[0] = tanh( W[0] * input + B[0] );
	for (int i=1; i<nlayers-2; i++) H[i] = tanh(W[i] * H[i-1] + B[i]);
	return W[nlayers-2] * H[nlayers-3] + B[nlayers-2];
}

// Evaluate error and gradients on the data
// uses batch backpropagation
MatrixType MLP::evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what)
{
	assert(data.rows() == input_dim);
	assert(target.rows() == output_dim);
	int nmap = data.cols();
	assert(target.cols() == nmap);

	// Init values
	MatrixType error = 0.0;
	if (what & OptimizeTarget::EVALUATE_GRADIENT) gradients.fill(0.0);
	Vector delta[nlayers - 1];

	// Loop through all mappings
	for (int i=1; i<=nmap; i++) {
		// linear output function => straightforward calculation of dout
		delta[nlayers - 2] = forward(data(all,i)) - target(all,i);
		// Add this to the error while we are at it
		if (what & OptimizeTarget::EVALUATE_ERROR)
			error += dot(delta[nlayers - 2],delta[nlayers - 2]);
		// Do gradients only if necessary
		if (what & OptimizeTarget::EVALUATE_GRADIENT) {
			// backpropagation in hidden units: tanh' = 1 - tanh^2
			for (int l = nlayers-3; l>=0; l--)
				delta[l] = multiply(
					W[l+1].transpose() * delta[l+1],
					1.0 - multiply(H[l], H[l]) );
			// Evaluate the gradients
			gW[0] += delta[0] * data(all,i).transpose();
			gB[0] += delta[0];
			for (int l = 1; l<nlayers-1; l++) {
				gW[l] += delta[l] * H[l-1].transpose();
				gB[l] += delta[l];
			}
		}
	}
	// Weight decay terms
	if (what & OptimizeTarget::EVALUATE_GRADIENT)
		for (int l = 0; l<nlayers-1; l++) gW[l] += prior * W[l];
	if (what & OptimizeTarget::EVALUATE_ERROR)
		error += prior * dot(values(1, weight_dim), values(1, weight_dim));

	// Do not average (not necessary)
//	gradients /= nmap;
//	error /= nmap;

	return 0.5 * error; // half the sum of square, for the derivative...
}


class LagMLP2 : public MLP
{
protected:
	int lag;
	// Evaluate error and gradients on the data
	// uses batch backpropagation
	MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH);
public:
	LagMLP2(int _lag, double prior, int nin, int nhidden, int nout)
	: MLP(3, prior, nin*_lag, nhidden, nout), lag(_lag) {}
};

// Evaluate error and gradients on the data
// uses batch backpropagation
// Redefined to use a lag vector on the data
MatrixType LagMLP2::evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what)
{
	assert(data.rows()*lag == input_dim);
	assert(target.rows() == output_dim);
	int nmap = data.cols() - lag+1;
	assert(target.cols() == nmap);

	// Init values
	MatrixType error = 0.0;
	if (what & OptimizeTarget::EVALUATE_GRADIENT) gradients.fill(0.0);
	Vector delta[nlayers - 1];

	// Loop through all mappings
	for (int i=1; i<=nmap; i++) {
		// linear output function => straightforward calculation of dout
		delta[nlayers - 2] = forward(data(all,i,i+lag-1).reshape(input_dim,1))
			- target(all,i);
		// Add this to the error while we are at it
		if (what & OptimizeTarget::EVALUATE_ERROR)
			error += dot(delta[nlayers - 2],delta[nlayers - 2]);
		// Do gradients only if necessary
		if (what & OptimizeTarget::EVALUATE_GRADIENT) {
			// backpropagation in hidden units: tanh' = 1 - tanh^2
			for (int l = nlayers-3; l>=0; l--)
				delta[l] = multiply(
					W[l+1].transpose() * delta[l+1],
					1.0 - multiply(H[l], H[l]) );
			// Evaluate the gradients
			gW[0] += delta[0] * data(all,i,i+lag-1).reshape(1,input_dim);
			gB[0] += delta[0];
			for (int l = 1; l<nlayers-1; l++) {
				gW[l] += delta[l] * H[l-1].transpose();
				gB[l] += delta[l];
			}
		}
	}
	// Weight decay terms
	if (what & OptimizeTarget::EVALUATE_GRADIENT)
		for (int l = 0; l<nlayers-1; l++) gW[l] += prior * W[l];
	if (what & OptimizeTarget::EVALUATE_ERROR)
		error += prior * dot(values(1, weight_dim), values(1, weight_dim));

	return 0.5 * error; // half the sum of square, for the derivative...
}

int help(int argc, char** argv)
{
	cout <<
"Train a 2 layer MLP network.\n"
"  Usage: " << argv[0] << " lag nhidden prior maxiter max_step_ahead [sanity] [logEOF...]\n"
"\n"
". lag is the length of the lag vector, in days.\n"
". nhidden is the number of neurons in the hidden layer.\n"
". prior is a multiplicative constant for the weight decay. It can "
"be interpreted as the inverse of the covariance presumed on the weights. "
"A value of 0 will cancel weight decay.\n"
". maxiter is the maximum number of iterations for the training phase. "
"It can be 0, in which case the prediction errors will be computed directly.\n"
". max_step_ahead is the maximum number of days ahead for which the "
"prediction RMSE will be computed.\n"
". 'sanity' is an optional argument that will force a check of the "
"batch backpopagation algorithm.\n"
". logEOF... is a list of numbers. The predicted series associated with "
"the corresponding EOFs will be logged.\n"
"\n"
"If possible the network will be loaded before the training, and saved "
"after each successful step, in the file " << netFileName << ". This also "
"means you can interrupt the program and resume computations later.\n"
	;
	return 1;
}
#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
	cerr << msg << " " << arg << endl; \
	return 2; \
}

int main(int argc, char** argv)
{
	time_t beginTime = time(0);
	// we stay in the background when losing connection to the terminal
	signal(SIGHUP, SIG_IGN); // not necessary with zsh, but for the others...

	const int mandatory_args = 5;
	if (argc<=mandatory_args) return help(argc, argv);
	process_arg(int, lag, atoi, argv[1], lag<1, "Invalid lag")
	process_arg(int, nhidden, atoi, argv[2], nhidden<1, "Invalid number of hidden units")
	process_arg(double, prior, atof, argv[3], prior<0, "Invalid prior")
	process_arg(int, maxiter, atoi, argv[4], maxiter<0, "Invalid max limit for iterations")
	process_arg(int, max_step_ahead, atoi, argv[5], max_step_ahead<1, "Invalid max step ahead prediction limit")

	randinit(seed); // have reproducible results

	ofstream output;
	output.precision(10);
	output.setf(ios::scientific);

	cout << "Loading data." << endl;
	Matrix training, validation, eofVectors;
	Vector std_correction, eofSingulars;
	if (!load(eofVectors,eofVectorsFileName)) {
		cerr << "Could not load eofVectors from '" << eofVectorsFileName << "'." << endl;
		return 2;
	} else cout << "eofVectors loaded from '" << eofVectorsFileName << "'." << endl;
	if (!load(eofSingulars,eofSingularsFileName)) {
		cerr << "Could not load eofSingulars from '" << eofSingularsFileName << "'." << endl;
		return 2;
	} else cout << "eofSingulars loaded from '" << eofSingularsFileName << "'." << endl;
	const int neof = eofVectors.cols();
	if (!neof || neof != eofSingulars.rows()) {
		cerr << "Invalid number of EOFs in eofVectors or eofSingulars." << endl;
		return 2;
	}
	const int m = eofVectors.rows();
	cout << "Data dimension is " << m << ", number of variables=" << nvar << ", number of EOFs=" << neof << endl;
	if (!load(std_correction, std_correctionFileName)) {
		cerr << "Could not std_correction from '" << std_correctionFileName << "'." << endl;
		return 2;
	} else cout << "std_correction loaded from '" << std_correctionFileName << "'." << endl;
	if (m != std_correction.rows()) {
		cerr << "std_correction dimension does not match EOFs dimension." << endl;
		return 2;
	}
	if (!load(training, trainingFileName)) {
		cerr << "Could not training from '" << trainingFileName << "'." << endl;
		return 2;
	} else cout << "training loaded from '" << trainingFileName << "'." << endl;
	if (neof != training.rows()) {
		cerr << "Training data dimension does not match number of EOFs." << endl;
		return 2;
	}
	const int t = training.cols();
	cout << "Training data size is " << t << " elements."<< endl;
    if (!load(validation, validationFileName)) {
		cerr << "Could not validation from '" << validationFileName << "'." << endl;
		return 2;
	} else cout << "validation loaded from '" << validationFileName << "'." << endl;
	if (neof != validation.rows()) {
		cerr << "Validation data dimension does not match number of EOFs." << endl;
		return 2;
	}
	const int v = validation.cols();
	cout << "Validation data size is " << v << " elements."<< endl;
	const int var_size = m / nvar;

	int nlogeof = 0;
	int *loggedeofnum = 0;
	if (argc>mandatory_args+1 && strcmp(argv[mandatory_args+1],"sanity")) {
		loggedeofnum = new int[argc-mandatory_args-1];
		for (int i=mandatory_args+1; i<argc; i++) {
			int num = atoi(argv[i]);
			if (num <= 0 || num > neof) {
				cerr << "Skipping invalid EOF number " << argv[i] << endl;
				continue;
			} else {
				cout << "EOF number " << num << " will be logged." << endl;
			}
			loggedeofnum[nlogeof++] = num;
		}
	}

	cout << "Creating " << neof*lag << "/" << nhidden << "/" << neof << " MLP Network (lag=" << lag << ") with prior=" << prior << "." <<endl;
	// Create the mlp
	LagMLP2 mlp(lag, prior, neof, nhidden, neof);

	if (argc>mandatory_args+1 && !strcmp(argv[mandatory_args+1],"sanity")) {
		cout << "Sanity check... " << flush;
		output.open("sanity_check.txt");
		Matrix check = mlp.gradcheck(training(all,1,t-1),training(all,lag+1,t));
		output << check << endl;
		output.close();
		cout << "results in 'sanity_check.txt'." << endl;
		return 0;
	}

	SCG optimizer;
	optimizer.option(1) = maxiter;
	optimizer.log(&cout);
	optimizer.autosave(netFileName);

	// Try to load the network first
	if (mlp.load(netFileName))
		cout << "Network loaded from '" << netFileName << "'." << endl;
	else cout << "Could not load network from '" << netFileName << "'. Using a random one." << endl;
	// Train it
	cout << "Training Network" <<endl;
	optimizer.optimize(mlp,training(all,1,t-1),training(all,lag+1,t));
	// Save it for the last time
	if (mlp.save(netFileName))
		cout << "Network saved in '" << netFileName << "'." << endl;
	else cout << "Could not save network in '" << netFileName << "'." << endl;

	// Test on the validation set
	cout << "Testing Network" <<endl;
	// Prepare a few log files
	ofstream logfile[nlogeof];
	char varfile[100];
	for (int l = 0; l<nlogeof; l++) {
		sprintf(varfile,"mlp_forecast_eof%d_predicted_series.txt",loggedeofnum[l]);
		logfile[l].open(varfile);
		logfile[l].precision(10);
		logfile[l].setf(ios::scientific);
	}
	Matrix error = zeros(m,max_step_ahead);
	Matrix valid_block(neof,lag);
	for (int i=1; i<=v-lag-max_step_ahead+1; i++) {
		valid_block.copy(validation(all,i,i+lag-1));
		for (int s=1; s<=max_step_ahead; s++) {
			Vector prediction = mlp.forward(valid_block.reshape(neof*lag,1));
			for (int l = 0; l<nlogeof; l++)
				logfile[l] << prediction(loggedeofnum[l]) << " ";
			// Reconstruct the data from the EOFs
			Vector eof_error = prediction - validation(all,i+lag+s-1);
			Vector rec_error = multiply(
				eofVectors * multiply(eofSingulars, eof_error),
				std_correction);
			// Add this to the total error
			error(all,s) += multiply(rec_error,rec_error);
			// update valid_block with the new value (if necessary)
			if (s<max_step_ahead) {
				valid_block(all,1,lag-1).copy(valid_block(all,2,lag));
				valid_block(all,lag).copy(prediction);
			}
		}
		// Close log files
		for (int l = 0; l<nlogeof; l++) {
			logfile[l] << endl;
			logfile[l].close();
		}
	}
	// Error is MSE for now. Have to average on variables _before_ taking
	// the square root.
	error /= v-lag-max_step_ahead+1;

	// RMSE calculation
	cout << "Calculating RMSE" <<endl;
	for (int var = 0; var < nvar; var++) {
		double rmse = 0;
		cout << "Variable " << var+1 << ": " << endl;
		char varfile[100];
		sprintf(varfile,"mlp_forecast_RMSE_var%d.txt",var+1);
		output.open(varfile);
		for (int s=1; s<=max_step_ahead; s++) {
			rmse = sqrt( mean(error(var*var_size+1, (var+1)*var_size, s, s)) );
			cout << s << " step ahead: " << rmse << endl;
			output << rmse << " " << endl;
		}
		output << endl; output.close();
	}

	cout << "Total time " << time(0) - beginTime << " seconds." << endl;
	return 0;
}

