/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdarg.h>
#include "CM_util.h"
#include "CM_optimize.h"
#include "CM_lapack.h"

const string dataFileName = "data.data";
const string meanFileName = "mean.data";
const string stdevFileName = "stdev.data";
const int t=1826, v = 1520, nvar = 1;
const int BUF_SIZE = 1024*1024*4;
const long seed = 42;
const string netFileName = "adjoint_network.values";
const string netFileName2 = "direct_network.values";

using namespace CheapMatrix;

class DirectPIP; // forward declaration

// Decouple the PIP problem: AdjointPIP uses the adjoint patterns matrix
// as weights for the first layer. The error function corresponds to
// finding the best subspace and MLP that minimize the prediction
// error _in_that_subspace_ (doesn't necessary tell anything about
// the error in the original data space).

class AdjointPIP : public OptimizeTarget {
friend class DirectPIP;
protected:
    Matrix DTM;  // adjoint patterns matrix
    Matrix W, V; // MLP weights
    Vector C, D; // MLP hidden and output layer biases
    Matrix gDTM;   // gradients - adjoint patterns matrix
    Matrix gW, gV; // gradients - MLP weights
    Vector gC, gD; // gradients - MLP hidden and output layer biases
    Vector H;   // Neuron values for the mlp hidden layer
    int n, m, nhidden; // data, reduced and hidden dimensions
    int lag, mlag;        // lag size, m*lag
    int total_dim, weight_dim;
    // Evaluate error and gradients on the data
    MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH);
    double prior;
public:
    AdjointPIP(int _lag, int _ninputs, int _npips, int _nhidden, double _prior);
    virtual ~AdjointPIP();
    // Projects input data. Returns the projected data
    virtual Matrix projection(const Matrix& input);
    virtual Matrix adjoint_patterns();
    // Forwards projected data in the form of a lag vector
    // Returns the output vector
    virtual Vector mlp_forward(const Vector& projected);
};

AdjointPIP::AdjointPIP(int _lag, int _ninputs, int _npips, int _nhidden, double _prior)
: lag(_lag), n(_ninputs), m(_npips), nhidden(_nhidden), prior(_prior)
{
    mlag = m*lag;
    int lim1 = m*n;                        // the adjoint patterns
    int lim2 = lim1 + nhidden*mlag;       // 1st layer weights
    int lim3 = lim2 + m*nhidden;           // 2nd layer weights
    int lim4 = lim3 + nhidden;             // hidden biases
    total_dim = lim4;// + m;                  // output biases
    weight_dim = lim3;
    values = randn(total_dim);    // Set initial values to something random
    gradients = zeros(total_dim); // And gradients to 0 (doesn't matter)
    // Set the references right
    DTM = values(1, lim1).reshape(m,n);
    // normalize DTM to begin with
    Vector norms = sqrt(sum(multiply(DTM,DTM),2));
    for (int i=1; i<=n; i++) DTM(all,i).divide(norms);
//    DTM /= sqrt(n + 1);
    W = values(lim1+1, lim2).reshape(nhidden,mlag);
    W /= sqrt(mlag + 1);
    V = values(lim2+1, lim3).reshape(m,nhidden);
    V /= sqrt(nhidden + 1);
    C = values(lim3+1, lim4);
    C /= sqrt(nhidden + 1);
//    D = values(lim4+1, total_dim);
//    D /= sqrt(m + 1);
    gDTM = gradients(1, lim1).reshape(m,n);
    gW = gradients(lim1+1, lim2).reshape(nhidden,mlag);
    gV = gradients(lim2+1, lim3).reshape(m,nhidden);
    gC = gradients(lim3+1, lim4);
//    gD = gradients(lim4+1, total_dim);
}

AdjointPIP::~AdjointPIP()
{
}

// Projects input data. Returns the projected data
Matrix AdjointPIP::projection(const Matrix& input)
{
    return DTM * input;
}

Matrix AdjointPIP::adjoint_patterns() {
    return DTM;
}

// Forwards projected data
// Returns the output vector
Vector AdjointPIP::mlp_forward(const Vector& Zlag)
{
    H = tanh( W * Zlag + C );
    return V * H;// + D;
}

// Evaluate error and gradients on the data
// uses batch backpropagation
// Don't use the target, and require the whole set as argument
// This is to optimize the projections...
MatrixType AdjointPIP::evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what)
{
    assert(data.rows() == n);
    int nmap = data.cols() - lag;

    // Project all the data first
    Matrix projected = projection(data);

    // Init values
    MatrixType error = 0.0;
    if (what & OptimizeTarget::EVALUATE_GRADIENT) gradients.fill(0.0);

    // Loop through all mappings
    for (int i=1; i<=nmap; i++) {
        // linear output function => straightforward calculation of dout
        Vector dout = mlp_forward(projected(all,i,i+lag-1).reshape(mlag,1))
            - projected(all,i+lag);
        // Add this to the error while we are at it
//        if (what & OptimizeTarget::EVALUATE_ERROR) error += dot(dout,dout);
        error += dot(dout,dout);
        // Do gradients only if necessary
        if (what & OptimizeTarget::EVALUATE_GRADIENT) {
            // backpropagation in hidden units: tanh' = 1 - tanh^2
            Vector dh = multiply(V.transpose() * dout, 1.0 - multiply(H, H));
            // backpropagation in projected subspace
            Vector dz = W.transpose() * dh; // linear transfert function
            // Evaluate the gradients
            // have to be careful because gDTM appears multiple times with
            // the lag vectors (derivative formulae) => add the terms
            // and dz is size m*lag, so split it
            for (int l=0; l<lag; l++)
                gDTM += dz(l*m+1,(l+1)*m) * data(all,i+l).transpose();
            // Ah! there is also another term...
            gDTM -= dout * data(all,i+lag).transpose();
            gW += dh * projected(all,i,i+lag-1).reshape(1,mlag);
            gC += dh;
            gV += dout * H.transpose();
//            gD += dout;
        }
    }
    // penalty term to force adjoint patterns to have norm 1
    Vector penalty = sum(multiply(DTM,DTM),2)-1;
    // variance normalization term
    Matrix normproj = sum(multiply(projected,projected),1);
    MatrixType sumnp = sum(normproj,2);
    // Weight decay terms
    if (what & OptimizeTarget::EVALUATE_GRADIENT) {
        Matrix g2 = ((-2.0*error) * sum( pow(normproj, -1.5) )) * projected * data.transpose();
        gDTM = g2 + gDTM / sumnp;
        for (int i=1; i<=n; i++)
            gDTM(all,i) += 2 * multiply(penalty,DTM(all,i));
        gW += prior * W;
        gV += prior * V;
    }
    error = error / sumnp;
//    if (what & OptimizeTarget::EVALUATE_ERROR) {
        error += dot(penalty, penalty);
        // weight decay term
        error += prior * dot(values(m*n+1, weight_dim), values(m*n+1, weight_dim));
//    }

    return 0.5 * error; // half the sum of square, for the derivative...
}

// DirectPIP does the opposite, using the PIP matrix as weights
// for a layer after the MLP. The error being computed in the data
// space, it minimizes the prediction on the data. But it is not
// as clear what happens in the subspace, requires a matrix inversion,
// and I think more adapted to on-line training.
class DirectPIP : public OptimizeTarget {
protected:
    Matrix W, V; // MLP weights
    Vector C;    // MLP hidden layer biases
    Matrix gW, gV; // gradients - MLP weights
    Vector gC;     // gradients - MLP hidden layer biases
    Vector H;     // Neuron values for the mlp hidden layer
    Matrix P;    // patterns matrix
    Matrix gP;   // gradients - patterns matrix
    int n, m, nhidden; // data, reduced and hidden dimensions
    int lag, mlag;        // lag size, m*lag
    int total_dim, weight_dim;
    // Evaluate error and gradients on the data
    MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH);
    double prior;
public:
    // create the complementary model
    DirectPIP(AdjointPIP adj);
    virtual ~DirectPIP();
    virtual Matrix projection(const Matrix& input);
    virtual Matrix patterns();
    // Forward the input data in the form of a lag vector
    // Return the predicted low-dim vector
    virtual Vector mlp_forward(const Vector& projected);
};

// create the complementary model
DirectPIP::DirectPIP(AdjointPIP adj)
{
    n = adj.n; m = adj.m; nhidden = adj.nhidden;
    lag = adj.lag; mlag = adj.mlag; total_dim = adj.total_dim;
    prior = adj.prior;
    int lim1 = nhidden*mlag;       // 1st layer weights
    int lim2 = lim1 + m*nhidden;   // 2nd layer weights
    int lim3 = lim2 + nhidden;     // hidden biases
    int lim4 = lim3 + m*n;         // the direct patterns
    assert(total_dim == lim4);
    weight_dim = lim2;
    values = randn(total_dim);    // Set initial values to something random
    gradients = zeros(total_dim); // And gradients to 0 (doesn't matter)
    // Set the references right
    // Reuse existing values
    W = values(1, lim1).reshape(nhidden,mlag);
    W.copy(adj.W);
    V = values(lim1+1, lim2).reshape(m,nhidden);
    V.copy(adj.V);
    C = values(lim2+1, lim3);
    C.copy(adj.C);
    P = values(lim3+1, lim4).reshape(n,m);
    // Compute the pseudo-inverse
    // $ P = DTM^T * (DTM*DTM^T)^{-1} $
    Matrix Sym = adj.DTM * adj.DTM.transpose();
    if (!inverse(Sym)) {cerr<<"Degenerated adjoint patterns!"<<endl; exit(2);}
    P.copy(adj.DTM.transpose() * Sym);
    gW = gradients(1, lim1).reshape(nhidden,mlag);
    gV = gradients(lim1+1, lim2).reshape(m,nhidden);
    gC = gradients(lim2+1, lim3);
    gP = gradients(lim3+1, lim4).reshape(n,m);
}

DirectPIP::~DirectPIP()
{
}

Matrix DirectPIP::patterns() {
    return P;
}

// Project a full dataset, calculating the adjoint patterns
Matrix DirectPIP::projection(const Matrix& data)
{
    Matrix Sym = P.transpose() * P;
    if (!inverse(Sym)) {cerr<<"Degenerated case!"<<endl; exit(2);}
    return (P * Sym).transpose() * data;
}

// Forward the input data in the form of a lag vector
// Return the predicted low-dim vector
Vector DirectPIP::mlp_forward(const Vector& Zlag)
{
    H = tanh( W * Zlag + C );
    return V * H;
}

// Evaluate error and gradients on the data
// uses batch backpropagation
MatrixType DirectPIP::evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what)
{
    assert(data.rows() == m);
    assert(target.rows() == n);
    int nmap = data.cols() - lag + 1;
    assert(target.cols() == nmap);

    // Init values
    MatrixType error = 0.0;
    if (what & OptimizeTarget::EVALUATE_GRADIENT) gradients.fill(0.0);
    // Loop through all mappings
    for (int i=1; i<=nmap; i++) {
        // linear output function => straightforward calculation of dout
        Vector Z = mlp_forward(data(all,i,i+lag-1).reshape(mlag,1));
        Vector dout = P * Z - target(all,i);
        // Add this to the error while we are at it
        if (what & OptimizeTarget::EVALUATE_ERROR) error += dot(dout,dout);
        // Do gradients only if necessary
        if (what & OptimizeTarget::EVALUATE_GRADIENT) {
            // backpropagation in predicted projected subspace
            Vector dz = P.transpose() * dout; // linear transfert function
            // backpropagation in hidden units: tanh' = 1 - tanh^2
            Vector dh = multiply(V.transpose() * dz, 1.0 - multiply(H, H));
            // Evaluate the gradients
            gP += dout * Z.transpose();
            gV += dz * H.transpose();
            gW += dh * data(all,i,i+lag-1).reshape(1,mlag);
            gC += dh;
        }
    }
    // Weight decay terms
    if (what & OptimizeTarget::EVALUATE_GRADIENT) {
        gW += prior * W;
        gV += prior * V;
    }
    if (what & OptimizeTarget::EVALUATE_ERROR) {
        // weight decay term
        error += prior * dot(values(1, weight_dim), values(1, weight_dim));
    }

    return 0.5 * error; // half the sum of square, for the derivative...
}

int help(int argc, char** argv)
{
    cout <<
"  Usage: " << argv[0] << " lag npips nhidden prior maxiter maxepoch max_step_ahead [sanity] [logPIP...]\n"
"\n"
    ;
    return 1;
}
#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
    cerr << msg << " " << arg << endl; \
    return 2; \
}

int main(int argc, char** argv)
{
    time_t beginTime = time(0);
    // we stay in the background when losing connection to the terminal
    signal(SIGHUP, SIG_IGN); // not necessary with zsh, but for the others...

    const int mandatory_args = 7;
    if (argc<=mandatory_args) return help(argc, argv);
    process_arg(int, lag, atoi, argv[1], lag<1, "Invalid lag")
    process_arg(int, npips, atoi, argv[2], lag<1, "Invalid number of patterns")
    process_arg(int, nhidden, atoi, argv[3], nhidden<1, "Invalid number of hidden units")
    process_arg(double, prior, atof, argv[4], prior<0, "Invalid prior")
    process_arg(int, maxiter, atoi, argv[5], maxiter<0, "Invalid max limit for iterations")
    process_arg(int, maxepoch, atoi, argv[6], maxiter<0, "Invalid max limit for epochs")
    process_arg(int, max_step_ahead, atoi, argv[7], max_step_ahead<1, "Invalid max step ahead prediction limit")

    randinit(seed); // have reproducible results

    ofstream output;
    output.precision(10);
    output.setf(ios::scientific);

    Matrix data;
    cout<<"Loading data..."<<endl;
    if (!load(data,dataFileName)) {
        cerr << "Could not load data from '" << dataFileName << "'." << endl;
        return 2;
    } else cout << "Data loaded from '" << dataFileName << "'." << endl;
    const int n=data.cols(), m=data.rows();
    Vector mean_correction, stdev_correction;
    if (load(mean_correction, meanFileName))
        cout << "mean_correction loaded from '" << meanFileName << "'." << endl;
    else {cerr << "Could not load mean_correction!" << endl; return 2;}
    if (load(stdev_correction, stdevFileName))
        cout << "stdev_correction loaded from '" << stdevFileName << "'." << endl;
    else {cerr << "Could not load stdev_correction!" << endl; return 2;}
    cout << "Data correction." << endl;
    // Correct all the data (extand to the validation set as well)
    for (int i=1; i<=m; i++) data(i,all) -= mean_correction(i);
    for (int j=1; j<=n; j++) data(all,j).divide(stdev_correction);
    Matrix training = data(all,1,t);
    Matrix validation = data(all,t+1,n);
    assert(t+v==n);
    const int var_size = m / nvar;

    int nlogpip = 0;
    int *loggedpipnum = 0;
    if (argc>mandatory_args+1 && strcmp(argv[mandatory_args+1],"sanity")) {
        loggedpipnum = new int[argc-mandatory_args-1];
        for (int i=mandatory_args+1; i<argc; i++) {
            int num = atoi(argv[i]);
            if (num <= 0 || num > npips) {
                cerr << "Skipping invalid PIP number " << argv[i] << endl;
                continue;
            } else {
                cout << "PIP number " << num << " will be logged." << endl;
            }
            loggedpipnum[nlogpip++] = num;
        }
    }

    // Create the network
    AdjointPIP adjpip(lag, m, npips, nhidden, prior);

    if (argc>mandatory_args+1 && !strcmp(argv[mandatory_args+1],"sanity")) {
        cout << "Sanity check... " << flush;
        output.open("sanity_check.txt");
        Matrix check = adjpip.gradcheck(training,Matrix()); // no target
        output << check << endl;
        output.close();
        cout << "results saved in 'sanity_check.txt'." << endl;
        return 0;
    }

    SCG optimizer;
    optimizer.option(1) = maxiter;
    optimizer.log(&cout);
    optimizer.autosave(netFileName);

    // Try to load the network first
    if (adjpip.load(netFileName))
        cout << "Network loaded from '" << netFileName << "'." << endl;
    else {
        cout << "Could not load network from '" << netFileName << endl;
        // Try to start using the EOFs
        Matrix eofs;
        if (load(eofs,"eofVectors.data")) {
            cout<<"Starting from EOFs found in eofVectors.data"<<endl;
            adjpip.adjoint_patterns().copy(eofs.transpose());
        }
        else {
            cout<<"EOFs cannot be found in eofVectors.data"<<endl;
            cout<<"Using a random network." << endl;
        }
    }
    // Train it
    cout << "Training Network" <<endl;
    // Use null matrix as target, only data is used
    optimizer.optimize(adjpip,training,Matrix());
    // Save it for the last time
    if (adjpip.save(netFileName))
        cout << "Network saved in '" << netFileName << "'." << endl;
    else cout << "Could not save network in '" << netFileName << "'." << endl;

    cout << "Projecting validation data using adjoint patterns." << endl;
    Matrix validproj = adjpip.projection(validation);
    output.open("adjoint_PIP_projected_validation.txt");
    output << validproj << endl;
    output.close();
    cout << "Projection saved in 'adjoint_PIP_projected_validation.txt'." << endl;
    Matrix DTM = adjpip.adjoint_patterns();
    output.open("adjoint_PIP_adjoint_patterns.txt");
    output << DTM << endl;
    output.close();
    cout << "Adjoint patterns saved in 'adjoint_PIP_adjoint_patterns.txt'." << endl;

    cout << "Creating the complementary model." << endl;
    DirectPIP dirpip(adjpip);
//    cout << dirpip.gradcheck(dirpip.projection(training(all,1,99)),training(all,lag+1,100));

    if (maxiter==0) {
        cout << "There was no adjoint training. " <<endl;
        // Try to load the network first
        if (dirpip.load(netFileName2))
            cout << "Direct network loaded from '" << netFileName2 << "'." << endl;
        else {
            cout << "Could not load direct network from '" << netFileName2 << "'." << endl;
            cout << "Assuming adjoint network is already trained." << endl;
        }
    }

    // Train it
    cout << "Training Network in data space" <<endl;
    SCG scg2;
    scg2.option(1) = 1;
    for (int i=1; i<=maxepoch; i++) {
        cout << "Epoch " << i << ", computing error... " << flush;
        cout << scg2.optimize(dirpip,dirpip.projection(training(all,1,t-1)),training(all,lag+1,t)) << endl;
        dirpip.save(netFileName2);
    }

    cout << "Projecting validation data using the new direct patterns." << endl;
    validproj = dirpip.projection(validation);
    output.open("direct_PIP_projected_validation.txt");
    output << validproj << endl;
    output.close();
    cout << "Projection saved in 'direct_PIP_projected_validation.txt'." << endl;
    Matrix P = dirpip.patterns();
    output.open("direct_PIP_direct_patterns.txt");
    output << P << endl;
    output.close();
    cout << "Direct patterns saved in 'direct_PIP_direct_patterns.txt'." << endl;

    // Test on the validation set
    cout << "Testing Network" <<endl;
    // Prepare a few log files
    ofstream logfile[nlogpip];
    char varfile[100];
    for (int l = 0; l<nlogpip; l++) {
        sprintf(varfile,"pip_forecast_pip%d_predicted_series.txt",loggedpipnum[l]);
        logfile[l].open(varfile);
        logfile[l].precision(10);
        logfile[l].setf(ios::scientific);
    }
    Matrix error = zeros(m,max_step_ahead);
    Matrix valid_block(npips,lag);
    for (int i=1; i<=v-lag-max_step_ahead+1; i++) {
        valid_block.copy(validproj(all,i,i+lag-1));
        for (int s=1; s<=max_step_ahead; s++) {
            Vector prediction = dirpip.mlp_forward(valid_block.reshape(npips*lag,1));
            for (int l = 0; l<nlogpip; l++)
                logfile[l] << prediction(loggedpipnum[l]) << " ";
            Vector data_prediction = P * prediction;
            // Reconstruct the error in the original data space
            Vector rec_error = multiply(
                data_prediction - validation(all,i+lag+s-1),
                stdev_correction);
            // Add this to the total error
            error(all,s) += multiply(rec_error,rec_error);
            // update valid_block with the new value (if necessary)
            if (s<max_step_ahead) {
                valid_block(all,1,lag-1).copy(valid_block(all,2,lag));
                valid_block(all,lag).copy(prediction);
            }
        }
        for (int l = 0; l<nlogpip; l++) logfile[l] << endl;
        cout << "." << flush;
        if (!(i%10)) cout << " " << flush;
        if (!(i%50)) cout << i << " / " << v-lag-max_step_ahead+1 << endl;
    }
    // Close log files
    for (int l = 0; l<nlogpip; l++) {
        logfile[l] << endl;
        logfile[l].close();
    }
    cout << " " << v-lag-max_step_ahead+1 << " / " << v-lag-max_step_ahead+1 << endl;
    // Error is MSE for now. Have to average on variables _before_ taking
    // the square root.
    error /= v-lag-max_step_ahead+1;

    // RMSE calculation
    cout << "Calculating RMSE" <<endl;
    for (int var = 0; var < nvar; var++) {
        double rmse = 0;
        cout << "Variable " << var+1 << ": " << endl;
        char varfile[100];
        sprintf(varfile,"pips_forecast_RMSE_var%d.txt",var+1);
        output.open(varfile);
        for (int s=1; s<=max_step_ahead; s++) {
            rmse = sqrt( mean(error(var*var_size+1, (var+1)*var_size, s, s)) );
            cout << s << " step ahead: " << rmse << endl;
            output << rmse << " " << endl;
        }
        output << endl; output.close();
    }

    cout << "Total time " << time(0) - beginTime << " seconds." << endl;
    return 0;
}
