/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "CM_util.h"

const string dataFileName = "Z500_85_94_5degrees.data";
const string meanFileName = "mean.data";
const string sineFileName = "seasonal_sines.data";
const string stdevFileName = "stdev.data";
const string all_eofVectorsFileName = "all_eofVectors.data";
const string all_eofSingularsFileName = "all_eofSingulars.data";
const int m = 320, n = 3346, t = 1826, v = 1520, nvar = 1;
const int var_size = m / nvar;

using namespace CheapMatrix;

int help(int argc, char** argv)
{
	cout <<
"Computes reconstruction errors every 10 EOFs.\n"
	;
	return 1;
}

int main(int argc, char** argv)
{
	time_t beginTime = time(0);
	// Stay alive even if parent is killed (as for connection lost)
	signal(SIGHUP, SIG_IGN);

	if (argc>1) return help(argc, argv);

	Matrix data, eofVectors;
	Vector eofSingulars;
	if (!load(data,dataFileName)) {
		cerr << "Could not load data from '" << dataFileName << "'." << endl;
		return 2;
	} else cout << "Data loaded from '" << dataFileName << "'." << endl;
	const int n=data.cols(), m=data.rows();
	cout << "Data dimension is (" << m << "," << n << ")." << endl;
	if (!load(eofVectors,all_eofVectorsFileName)) {
		cerr << "Could not load eofVectors from '" << all_eofVectorsFileName << "'." << endl;
		return 2;
	} else cout << "eofVectors loaded from '" << all_eofVectorsFileName << "'." << endl;
	if (m != eofVectors.rows()) {
		cerr << "EOF and data dimensions ("<< eofVectors.rows() << " and " << m << ") do not match." << endl;
		return 2;
	}
	if (!load(eofSingulars,all_eofSingularsFileName)) {
		cerr << "Could not load eofSingulars from '" << all_eofSingularsFileName << "'." << endl;
		return 2;
	} else cout << "eofSingulars loaded from '" << all_eofSingularsFileName << "'." << endl;
	const int maxeof = eofVectors.cols();
	if (!maxeof || maxeof != eofSingulars.rows()) {
		cerr << "Invalid number of EOFs in eofVectors or eofSingulars." << endl;
		return 2;
	}
	cout << "Total number of EOFs=" << maxeof << endl;
	Vector mean_correction, stdev_correction;
	Matrix params;
	if (load(mean_correction, meanFileName))
		cout << "mean_correction loaded from '" << meanFileName << "'." << endl;
	else {cerr << "Could not load mean_correction!" << endl; return 2;}
	if (load(params, sineFileName))
		cout << "Seasonal sine parameters (amplitude, phase) loaded from '" << sineFileName << "'." << endl;
	else {cerr << "Could not load sine parameters!" << endl; return 2;}
	if (load(stdev_correction, stdevFileName))
		cout << "stdev_correction loaded from '" << stdevFileName << "'." << endl;
	else {cerr << "Could not load stdev_correction!" << endl; return 2;}

	// PI isn't present on all systems
	// I should really declare some useful constants (e, pi, Inf, NaN, eps...)
	// in the namespace
	#ifndef M_PI
	#define M_PI 3.14159265358979323846
	#endif
	const double omega = 2.0*M_PI/365.25;

	cout << "Data correction." << endl;
	// Correct all the data (extand to the validation set as well)
	for (int i=1; i<=m; i++) {
		data(i,all) -= params(1,i)*sin(omega*range(1,n).transpose()+params(2,i)) + mean_correction(i);
	}
	for (int j=1; j<=n; j++) data(all,j).divide(stdev_correction);

	// Prepare a few log files
	ofstream tlogfile[nvar];
	ofstream vlogfile[nvar];
	char varfile[100];
	for (int var = 0; var < nvar; var++) {
		sprintf(varfile,"reconstruction_RMSE_var%d_training.txt",var+1);
		tlogfile[var].open(varfile);
		tlogfile[var].precision(10);
		tlogfile[var].setf(ios::scientific);
		sprintf(varfile,"reconstruction_RMSE_var%d_validation.txt",var+1);
		vlogfile[var].open(varfile);
		vlogfile[var].precision(10);
		vlogfile[var].setf(ios::scientific);
	}

	cout << "Data projection and RMSE computations." << endl;
	for (int e = 10; e <= maxeof; e+=10) {
		// Use the fact that singular vectors are orthogonal with norm 1
		// And correct the result using the singular values:
		// A = U * S * VT   =>   VT = S^-1 * UT * A
		// Actually, don't divide to multiply just after by the same thing
		// And add () to force the second * before (lower dim = faster)
		// Do it column by column to save memory
		// training and validation error vectors
		Vector terr = zeros(m), verr = zeros(m);
		for (int j=1; j<=t; j++) {
			Vector recdiff = eofVectors(all,1,e) *
			                 (eofVectors(all,1,e).transpose() * data(all,j))
			                 - data(all,j);
			// stdev correction
			Vector rec_error = multiply(recdiff, stdev_correction);
			// Add this to the total error
			terr += multiply(rec_error,rec_error);
		}
		// Validation
		for (int j=t+1; j<=n; j++) {
			Vector recdiff = eofVectors(all,1,e) *
			                 (eofVectors(all,1,e).transpose() * data(all,j))
			                 - data(all,j);
			// stdev correction
			Vector rec_error = multiply(recdiff, stdev_correction);
			// Add this to the total error
			verr += multiply(rec_error,rec_error);
		}
		terr /= t; verr /= v; // mean for each component
		// Finish the averaging for each variable _before_ taking sqrt
		for (int var = 0; var < nvar; var++) {
			cout << "Using " << e << "EOFs, variable " << var+1 << ": " << flush;
			double trmse = sqrt( mean(terr(var*var_size+1, (var+1)*var_size)) );
			cout << "training RMSE = " << trmse << flush;
			tlogfile[var] << trmse << " " << flush;
			double vrmse = sqrt( mean(verr(var*var_size+1, (var+1)*var_size)) );
			cout << ", validation RMSE = " << vrmse << endl;
			vlogfile[var] << vrmse << " " << flush;
		}
	}

	// Close log files
	for (int var = 0; var < nvar; var++) {
		tlogfile[var] << endl;
		tlogfile[var].close();
		vlogfile[var] << endl;
		vlogfile[var].close();
	}

	cout << "Total time " << time(0) - beginTime << " seconds." << endl;
	return 0;
}
