/*
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program, see the file COPYING; if not, write
    to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.
*/

#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "CM_util.h"

const string dataFileName = "Z500_85_94_5degrees.data";
const string meanFileName = "mean.data";
const string sineFileName = "seasonal_sines.data";
const string stdevFileName = "stdev.data";
const string all_eofVectorsFileName = "all_eofVectors.data";
const string all_eofSingularsFileName = "all_eofSingulars.data";
const string eofVectorsFileName = "eofVectors.data";
const string eofSingularsFileName = "eofSingulars.data";
const string train_seriesFileName = "train.data";
const string valid_seriesFileName = "valid.data";
const int m = 320, n = 3346, t = 1826, v = 1520, nvar = 1;
const int var_size = m / nvar;
const int BUF_SIZE = 1024*1024;
const int seed = 42;

using namespace CheapMatrix;


int help(int argc, char** argv)
{
	cout <<
"Select the first EOFS and compute the associated series from the whole set.\n"
"  Usage: " << argv[0] << " num_EOF\n"
"\n"
	;
	return 1;
}
#define process_arg(type, var, func, arg, cond, msg) \
const type var = func(arg); \
if (cond) { \
	cerr << msg << " " << arg << endl; \
	return 2; \
}

int main(int argc, char** argv)
{
	time_t beginTime = time(0);
	// Stay alive if parent is killed (as for connection lost)
	signal(SIGHUP, SIG_IGN);

	const int mandatory_args = 1;
	if (argc<=mandatory_args) return help(argc, argv);
	process_arg(int, neof, atoi, argv[1], neof<1, "Invalid lag")

	Matrix data, eofVectors;
	Vector eofSingulars;
	if (!load(data,dataFileName)) {
		cerr << "Could not load data from '" << dataFileName << "'." << endl;
		return 2;
	} else cout << "Data loaded from '" << dataFileName << "'." << endl;
	const int n=data.cols(), m=data.rows();
	cout << "Data dimension is (" << m << "," << n << ")." << endl;
	if (!load(eofVectors,all_eofVectorsFileName)) {
		cerr << "Could not load eofVectors from '" << all_eofVectorsFileName << "'." << endl;
		return 2;
	} else cout << "eofVectors loaded from '" << all_eofVectorsFileName << "'." << endl;
	if (m != eofVectors.rows()) {
		cerr << "EOF and data dimensions ("<< eofVectors.rows() << " and " << m << ") do not match." << endl;
		return 2;
	}
	if (!load(eofSingulars,all_eofSingularsFileName)) {
		cerr << "Could not load eofSingulars from '" << all_eofSingularsFileName << "'." << endl;
		return 2;
	} else cout << "eofSingulars loaded from '" << all_eofSingularsFileName << "'." << endl;
	const int maxeof = eofVectors.cols();
	if (!maxeof || maxeof != eofSingulars.rows()) {
		cerr << "Invalid number of EOFs in eofVectors or eofSingulars." << endl;
		return 2;
	}
	cout << "Total number of EOFs=" << maxeof << endl;
	if (neof>maxeof) {
		cerr << "Cannot select " << neof << " out of " << maxeof << " EOFs." << endl;
		return 2;
	}
	Vector mean_correction, stdev_correction;
	Matrix params;
	if (load(mean_correction, meanFileName))
		cout << "mean_correction loaded from '" << meanFileName << "'." << endl;
	else {cerr << "Could not load mean_correction!" << endl; return 2;}
	if (load(params, sineFileName))
		cout << "Seasonal sine parameters (amplitude, phase) loaded from '" << sineFileName << "'." << endl;
	else {cerr << "Could not load sine parameters!" << endl; return 2;}
	if (load(stdev_correction, stdevFileName))
		cout << "stdev_correction loaded from '" << stdevFileName << "'." << endl;
	else {cerr << "Could not load stdev_correction!" << endl; return 2;}

	// PI isn't present on all systems
	// I should really declare some useful constants (e, pi, Inf, NaN, eps...)
	// in the namespace
	#ifndef M_PI
	#define M_PI 3.14159265358979323846
	#endif
	const double omega = 2.0*M_PI/365.25;

	cout << "Data correction." << endl;
	// Correct all the data (extand to the validation set as well)
	for (int i=1; i<=m; i++) {
		data(i,all) -= params(1,i)*sin(omega*range(1,n).transpose()+params(2,i)) + mean_correction(i);
	}
	for (int j=1; j<=n; j++) data(all,j).divide(stdev_correction);

	cout << "Data projection." << endl;
	// Use the fact that singular vectors are orthogonal with norm 1
	// And correct the result using the singular values:
	// A = U * S * VT   =>   VT = S^-1 * UT * A
	// Also, this multiplication doesn't mix validation & training so
	// can be done in one go on the data
	Matrix series = eofVectors(all,1,neof).transpose() * data;
	for (int j=1; j<=n; j++) series(all,j).divide( eofSingulars(1,neof) );

	if (save(eofVectors(all,1,neof), eofVectorsFileName))
		cout << "eofVectors saved in '" << eofVectorsFileName << "'." << endl;
	else {cerr << "Could not save eofVectors!" << endl; return 2;}
	if (save(eofSingulars(1,neof), eofSingularsFileName))
		cout << "eofSingulars saved in '" << eofSingularsFileName << "'." << endl;
	else {cerr << "Could not save eofSingulars!" << endl; return 2;}
	if (save(series(all,1,t), train_seriesFileName))
		cout << "training series saved in '" << train_seriesFileName << "'." << endl;
	else {cerr << "Could not save training series!" << endl; return 2;}
	if (save(series(all,t+1,n), valid_seriesFileName))
		cout << "validation series saved in '" << valid_seriesFileName << "'." << endl;
	else {cerr << "Could not save validation series!" << endl; return 2;}

	cout << "Total time " << time(0) - beginTime << " seconds." << endl;
	return 0;
}
